@@ -11,11 +11,11 @@ import time
 import random
 import re
 
+from django.db import connections
 from django.db.backends.base.base import BaseDatabaseWrapper
 from django.db.backends.base.validation import BaseDatabaseValidation
 from django.core.exceptions import ImproperlyConfigured
-
-from django.utils.six import binary_type, text_type
+from django.core import signals
 from django.utils.encoding import smart_str
 
 from .client import DatabaseClient
@@ -38,7 +38,7 @@ logger = logging.getLogger(__name__)
 def decoder(value, encodings=('utf-8',)):
     """This decoder tries multiple encodings before giving up"""
 
-    if not isinstance(value, binary_type):
+    if not isinstance(value, bytes):
         raise ValueError(f"Not a binary type: {value} {type(value)}")
 
     for enc in encodings:
@@ -55,7 +55,7 @@ class DatabaseWrapper(BaseDatabaseWrapper):
     Database = pyodbc
 
     DRIVER_MAP = {
-        'DARWIN': '/Applications/gbasedbt/lib/cli/iclit09b.dylib',
+        'DARWIN': '/Applications/GBS/gbasedbt/lib/cli/iclit09b.dylib',
         'LINUX': '/opt/gbase/lib/cli/iclit09b.so',
         'WINDOWS32bit': 'GBase ODBC DRIVER',
         'WINDOWS64bit': 'GBase ODBC DRIVER (64-bit)',
@@ -73,7 +73,7 @@ class DatabaseWrapper(BaseDatabaseWrapper):
         'BigAutoField': 'bigserial',
         'BinaryField': 'blob',
         'BooleanField': 'boolean',
-        'CharField': 'varchar(%(max_length)s)',
+        'CharField': 'lvarchar(%(max_length)s)',
         'CommaSeparatedIntegerField': 'lvarchar(%(max_length)s)',
         'DateField': 'date',
         'DateTimeField': 'datetime year to fraction(5)',
@@ -149,6 +149,10 @@ class DatabaseWrapper(BaseDatabaseWrapper):
 
         options = self.settings_dict.get('OPTIONS', {})
 
+        self._validation_enabled = options.get("VALIDATE_CONNECTION", False)
+        self._validation_interval = options.get("VALIDATION_INTERVAL", 300)
+        self._next_validation = time.time() + self._validation_interval
+        self._validation_query = options.get("VALIDATION_QUERY", "SELECT 1 FROM dual")
         self.encodings = options.get('encodings', ('utf-8', 'cp1252', 'iso-8859-1'))
         # make lookup operators to be collation-sensitive if needed
         self.collation = options.get('collation', None)
@@ -168,6 +172,26 @@ class DatabaseWrapper(BaseDatabaseWrapper):
         self.introspection = self.introspection_class(self)
         self.validation = self.validation_class(self)
 
+    def validate_connection(self):
+        """
+        This method is invoked at the start of a request to verify an existing
+        connection is still functional. This is achieved by doing a simple query
+        against the database.
+        """
+        if not self._validation_enabled or time.time() < self._next_validation:
+            return
+
+        self._next_validation = time.time() + self._validation_interval
+
+        # We call close_if_unusable_or_obsolete to ensure obsolete connections
+        # are closed before we consider validating them. This will result in
+        # close_if_unusable_or_obsolete being called twice since it is also
+        # called automatically by django. This is ok since the second call is
+        # essentially a no-op.
+        self.close_if_unusable_or_obsolete()
+        if self.connection is not None and not self.is_usable():
+            self.close()
+
     def get_driver_path(self):
         system = platform.system().upper()
         if system == 'WINDOWS':
@@ -193,9 +217,9 @@ class DatabaseWrapper(BaseDatabaseWrapper):
         if platform.system().upper() != 'WINDOWS':
             sqlhosts = os.environ.get('GBASEDBTSQLHOSTS')
             if not sqlhosts or not os.path.exists(sqlhosts):
-                raise ImproperlyConfigured('Cannot find GBasedbt sqlhosts at {}'.format(sqlhosts))
+                raise ImproperlyConfigured('Cannot find GBase 8s sqlhosts at {}'.format(sqlhosts))
             if not os.path.exists(options['DRIVER']):
-                raise ImproperlyConfigured('cannot find GBasedbt driver at {}'.format(options['DRIVER']))
+                raise ImproperlyConfigured('cannot find GBase 8s driver at {}'.format(options['DRIVER']))
         conn_params['OPTIONS'] = options
 
         conn_params['AUTOCOMMIT'] = False if 'AUTOCOMMIT' not in conn_params else conn_params['AUTOCOMMIT']
@@ -223,7 +247,7 @@ class DatabaseWrapper(BaseDatabaseWrapper):
             parts.append('CPTimeout={}'.format(conn_params['OPTIONS']['CPTIMEOUT']))
 
         connection_string = ';'.join(parts)
-        logging.debug('Connecting to GBasedbt')
+        logging.debug('Connecting to GBase 8s')
         self.connection = self._get_connection_with_retries(connection_string, conn_params)
         self.connection.setencoding(encoding='UTF-8')
 
@@ -243,6 +267,8 @@ class DatabaseWrapper(BaseDatabaseWrapper):
         self.connection.add_output_converter(-101, lambda r: r.decode('utf-8'))  # Constraints
         self.connection.add_output_converter(-391, lambda r: r.decode('utf-16-be'))  # Integrity Error
 
+        self.connection.add_output_converter(-114, self.handle_sql_infx_bigint_as_bigint)
+
         self.connection.add_output_converter(pyodbc.SQL_CHAR, self._output_converter)
         self.connection.add_output_converter(pyodbc.SQL_WCHAR, self._output_converter)
         self.connection.add_output_converter(pyodbc.SQL_VARCHAR, self._output_converter)
@@ -293,7 +319,7 @@ class DatabaseWrapper(BaseDatabaseWrapper):
 
     def _unescape(self, raw):
         """
-        For some reason the GBasedbt ODBC driver seems to double escape new line characters.
+        For some reason the GBase 8s ODBC driver seems to double escape new line characters.
 
         This little handler converts them back.
 
@@ -304,11 +330,14 @@ class DatabaseWrapper(BaseDatabaseWrapper):
     def _output_converter(self, raw):
         return decoder(self._unescape(raw), self.encodings)
 
+    def handle_sql_infx_bigint_as_bigint(self, value):
+        return int.from_bytes(value, byteorder='little', signed=False) 
+
     def init_connection_state(self):
         pass
 
     def create_cursor(self, name=None):
-        logging.debug('Creating GBasedbt cursor')
+        logging.debug('Creating GBase 8s cursor')
         return CursorWrapper(self.connection.cursor(), self)
 
     def _set_autocommit(self, autocommit):
@@ -331,13 +360,32 @@ class DatabaseWrapper(BaseDatabaseWrapper):
         self.cursor().execute(start_sql)
 
     def is_usable(self):
+        # We create a cursor and then explicitly close it as there is a bug
+        # that is encountered when relying on garbage collection to close the
+        # cursor: https://github.com/mkleehammer/pyodbc/issues/585
         try:
-            # Use a cursor directly, bypassing Django's utilities.
-            self.connection.cursor().execute("SELECT 1")
-        except pyodbc.Error:
+            cursor = self.connection.cursor()
+        except pyodbc.Error as exc:
+            logger.info(f"error creating cursor: {exc}")
             return False
-        else:
+
+        try:
+            cursor.execute(self._validation_query)
             return True
+        except pyodbc.Error as exc:
+            logger.info(f"error executing query: {exc}")
+            return False
+        finally:
+            # We close the cursor explicitly to work around the pyodbc bug
+            # described at the top of this function. If closing the cursor
+            # fails we set the return value to `False`. Otherwise it
+            # remains what was returned in the `try` or `except` block, which
+            # depends on whether `cursor.execute` succeeded or not.
+            try:
+                cursor.close()
+            except pyodbc.Error as exc:
+                logger.info(f"error closing cursor: {exc}")
+                return False
 
     def read_dirty(self):
         self.cursor().execute('set isolation to dirty read;')
@@ -401,7 +449,7 @@ class CursorWrapper(object):
             self.cursor.close()
 
     def format_sql(self, sql, params):
-        if isinstance(sql, text_type):
+        if isinstance(sql, str):
             # FreeTDS (and other ODBC drivers?) doesn't support Unicode
             # yet, so we need to encode the SQL clause itself in utf-8
             sql = smart_str(sql, self.driver_charset)
@@ -417,7 +465,7 @@ class CursorWrapper(object):
         fp = []
         if params is not None:
             for p in params:
-                if isinstance(p, text_type):
+                if isinstance(p, str):
                     if self.driver_charset:
                         # FreeTDS (and other ODBC drivers?) doesn't support Unicode
                         # yet, so we need to encode parameters in utf-8
@@ -425,7 +473,7 @@ class CursorWrapper(object):
                     else:
                         fp.append(p)
 
-                elif isinstance(p, binary_type):
+                elif isinstance(p, bytes):
                     fp.append(p)
 
                 elif isinstance(p, type(True)):
@@ -467,7 +515,7 @@ class CursorWrapper(object):
                 f = row[i]
                 # FreeTDS (and other ODBC drivers?) doesn't support Unicode
                 # yet, so we need to decode utf-8 data coming from the DB
-                if isinstance(f, binary_type):
+                if isinstance(f, bytes):
                     row[i] = f.decode(self.driver_charset)
         return tuple(row)
 
@@ -493,3 +541,12 @@ class CursorWrapper(object):
 
     def __iter__(self):
         return iter(self.cursor)
+
+
+def _validate_connection(**kwargs):
+    for conn in connections.all():
+        if isinstance(conn, DatabaseWrapper):
+            conn.validate_connection()
+
+
+signals.request_started.connect(_validate_connection)
